@@ -363,8 +363,9 @@ static char recv_packet(Packet packet, io_buffer fp, Dword maxlength)
 }
 
 static ssize_t read_packet(Packet packet, int fd, Dword length) {
+	ssize_t ret;
 	make_new(packet, length);
-	ssize_t ret = read(fd, packet->data, packet->len);
+	ret = read(fd, packet->data, packet->len);
 	if(ret < 0)
 		return ret;
 	packet->len = (Dword) ret;
@@ -559,14 +560,14 @@ static uint16_t getportnum(char *portnum)
 
 	for (port = 0; isdigit(*digits); ++digits)
 		{
-			port = (port * 10) + (*digits - '0');
+			port = (port * 10) + (uint8_t)(*digits - '0');
 		}
 
 	if ((*digits != '\0') || (port <= 0))
 		{
 			if ((serv = getservbyname(portnum, "tcp")) != NULL)
 				{
-					port = ntohs(serv->s_port);
+					port = ntohs((uint16_t)serv->s_port);
 				}
 			else
 				{
@@ -750,7 +751,7 @@ static void alarm_signal(int a UNUSEDP)
  * This is the main loop when running as a server.
  */
 static void server_main_loop(int sock, char **device_name,
-			     unsigned  int n_dev) NORETURN;
+			     unsigned int n_dev) NORETURN;
 static void server_main_loop(int sock, char **device_name,
 			     unsigned int n_dev)
 {
@@ -884,7 +885,7 @@ int main (int argc, char** argv)
 
 	if(optind < argc) {
 		device_name = argv + optind;
-		n_dev = argc - optind;
+		n_dev = (unsigned int) (argc - optind);
 	} else {
 		device_name = (char **)&floppy0;
 		n_dev = 1;
@@ -1035,7 +1036,7 @@ static void send_reply64(int rval, io_buffer sock, mt_off_t len) {
 	Packet reply = newPacket();
 
 	make_new(reply, 12);
-	put_qword(reply, 0, len);
+	put_qword(reply, 0, (Qword) len);
 	if (rval == -1) {
 		put_dword(reply, 8, 0);
 	} else {
@@ -1148,7 +1149,20 @@ void serve_client(int sockhandle, char **device_name, unsigned int n_dev,
 /*		if(opcode->data[0] != OP_CLOSE)*/
 		    recv_packet(parm, sock, MAX_DATA_REQUEST);
 
+		/* on its own, floppyd does several small writes,
+		 * running into the performance issue described in
+		 * https://eklitzke.org/the-caveats-of-tcp-nodelay
+		 * Cork fixes this by stalling the small writes until
+		 * floppyd has assembled its entire message, thus
+		 * preventing the bad interaction between Nagle's
+		 * algorithm and Linux' delayed ACKs. Another fix
+		 * would be to not send the small batches immediately,
+		 * but instead keep them around and submit them to the
+		 * kernel in one go using writev. However, writev is
+		 * not available everywhere
+		 */
 		cork(sock->handle, 1);
+
 		switch(opcode->data[0]) {
 			case OP_OPRO:
 				if(get_length(parm) >= 4)
@@ -1214,7 +1228,7 @@ void serve_client(int sockhandle, char **device_name, unsigned int n_dev,
 				} else {
 					rval = write_packet(parm, devFd);
 				}
-				send_reply(devFd, sock, rval);
+				send_reply(devFd, sock, (Dword) rval);
 				break;
 			case OP_SEEK:
 #if DEBUG
@@ -1222,7 +1236,8 @@ void serve_client(int sockhandle, char **device_name, unsigned int n_dev,
 #endif
 
 				lseek(devFd, 
-				      get_dword(parm, 0), get_dword(parm, 4));
+				      get_dword(parm, 0),
+				      (int) get_dword(parm, 4));
 				send_reply(devFd, 
 					   sock, 
 					   (Dword) lseek(devFd, 0, SEEK_CUR));
@@ -1240,7 +1255,8 @@ void serve_client(int sockhandle, char **device_name, unsigned int n_dev,
 				fprintf(stderr, "SEEK64:\n");
 #endif
 				mt_lseek(devFd, 
-					 get_qword(parm,0), get_dword(parm,8));
+					 (mt_off_t) get_qword(parm,0),
+					 (int) get_dword(parm,8));
 				send_reply64(devFd, 
 					     sock, 
 					     mt_lseek(devFd, 0, SEEK_CUR));
